/* [KSC4]-------------------------------------------------------------------[]
 *
 *  4X5X7POC.C - Multiplexing 4 5x7 LED Matrix Displays, version 0.01
 *               By Marco "Kiko" Carnut <kiko \at\ postcogito \dot\ org>
 *
 *  Description: This is a proof-of-concept implementation of multiplexing
 *  ------------ scheme for a single 20-pin ATtiny2313 to control four 5x7
 *  LED matrix displays. The code implements the basic display system, a
 *  brightness calibration procedure, a pixel control test and a classic
 *  scrolling message application.
 *
 *  The hardware basically consists of the AVR's I/O ports connected directly
 *  to the LED arrays without current limiting resistors. This means we can't
 *  have any LEDs turned on for too long or they'll likely blow up. As a
 *  precaution against software bugs that may cause that, we run with the
 *  watchdog timer enabled and set to the minimum period of 16ms, even though
 *  I'm not sure this is fast enough to prevent the LED from blowing up.
 *
 *  See also: http://www.postcogito.org/Kiko/LedMatrixDisplayPOC
 *  ---------
 *
 *  Portability: Tested on an ATtiny2313. Compiled under avr-gcc 3.4.4 and
 *  ------------ avr-libc 1.4.0.
 *
 *  License: This program is made available under the Creative Commons
 *  -------- Attribution-NonCommercial-ShareAlike version 2.5 license.
 *  Please see http://creativecommons.org/licenses/by-nc-sa/2.5/ for details.
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * []-----------------------------------------------------------------------[]
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>

#define F_CPU                          8000000L

#include <util/delay.h>

/* -------------------- Software Configuration Defines -------------------- */

#define HAVE_FRAME_FROM_TIMER
#define SCROLL_MESSAGE

/* --------------------------- Utility Defines ---------------------------- */

#define FLAG_HAS_INTERRUPT             0
#define FLAG_HAS_FRAME                 1
#define FLAG_SINGLE_STEP               2
#define FLAG_SWITCH_FRAME              3

/* ------------------ Global Variables and Declarations ------------------- */

register unsigned char cycle asm("r2");

unsigned char bitmap1[20],bitmap2[20];
register unsigned char *current asm("r4");
register unsigned char *bitmap  asm("r6");

#ifndef HAVE_FRAME_FROM_TIMER

unsigned char frame;

#endif // HAVE_FRAME_FROM_TIMER

/* 4x7 Font definition. Each byte is divided in two nibbles, each represeting
 * the scanlines of two characters. As there are five rows, each two characters
 * take five bytes. Here we have ASCII characters 32 to 95.
 *
 * This was generated by the attached cvtfont.pl script from the font3x5.txt
 * file.
 */

unsigned char font[] PROGMEM = {
    0x04, 0x04, 0x04, 0x00, 0x04, 0xAA, 0xAE, 0x0A, 0x0E, 0x0A, 0x4A, 0x62,
    0xC4, 0x68, 0xCA, 0x44, 0xA4, 0x40, 0x80, 0x60, 0x28, 0x44, 0x44, 0x44,
    0x28, 0xA0, 0x44, 0xEE, 0x44, 0xA0, 0x00, 0x00, 0x0E, 0x40, 0x80, 0x02,
    0x02, 0x04, 0x08, 0x48, 0x44, 0xAC, 0xA4, 0xA4, 0x4E, 0xCC, 0x22, 0x44,
    0x82, 0xEC, 0xAE, 0xA8, 0xEC, 0x22, 0x2C, 0x6E, 0x82, 0xC4, 0xA4, 0x44,
    0x44, 0xAA, 0x46, 0xA2, 0x4C, 0x00, 0x44, 0x00, 0x44, 0x08, 0x20, 0x4E,
    0x80, 0x4E, 0x20, 0x8C, 0x42, 0x24, 0x40, 0x84, 0x04, 0x0A, 0x4E, 0xAA,
    0x6A, 0xC6, 0xA8, 0xC8, 0xA8, 0xC6, 0xCE, 0xA8, 0xAC, 0xA8, 0xCE, 0xE6,
    0x88, 0xCA, 0x8A, 0x86, 0xAE, 0xA4, 0xE4, 0xA4, 0xAE, 0x2A, 0x2A, 0x2C,
    0x2A, 0xCA, 0x8A, 0x8E, 0x8E, 0x8A, 0xEA, 0xEE, 0xAA, 0xAA, 0xAA, 0xAE,
    0xCE, 0xAA, 0xCA, 0x8C, 0x82, 0xC6, 0xA8, 0xC4, 0xA2, 0xAC, 0xEA, 0x4A,
    0x4A, 0x4A, 0x4E, 0xAA, 0xAA, 0xAE, 0xAE, 0x4A, 0xAA, 0xAA, 0x44, 0xA4,
    0xA4, 0xE6, 0x24, 0x44, 0x84, 0xE6, 0x8C, 0x84, 0x44, 0x24, 0x2C, 0x40,
    0xA0, 0x00, 0x00, 0x0E, 0x40, 0x2C, 0x06, 0x0A, 0x06, 0x80, 0xC6, 0xA8,
    0xA8, 0xC6, 0x20, 0x6C, 0xAA, 0xAC, 0x66, 0x60, 0x86, 0xCA, 0x86, 0x8C,
    0x84, 0x80, 0xCC, 0xA4, 0xAE, 0x48, 0x0A, 0x4C, 0x4A, 0x8A, 0x40, 0x4C,
    0x4E, 0x4E, 0x4A, 0x00, 0xC4, 0xAA, 0xAA, 0xA4, 0x00, 0xC6, 0xAA, 0xC6,
    0x82, 0x00, 0x66, 0x8C, 0x86, 0x8C, 0x80, 0xEA, 0x8A, 0x8A, 0x6E, 0x00,
    0xAA, 0xAE, 0xAE, 0x46, 0x00, 0xAA, 0x46, 0xA2, 0xAC, 0x06, 0xE4, 0x48,
    0x84, 0xE6, 0x4C, 0x44, 0x02, 0x44, 0x4C, 0xE0, 0x00, 0x00, 0x00, 0x00
};

/* DDRD and DDRD base values for multiplexing */

unsigned char b_ors[] EEMEM = { 
    0x20,0x08,0x00,0x00,0x00,0x20,0x08,0x00,0x00,0x00,
    0x40,0x80,0x00,0x00,0x00,0x40,0x80,0x00,0x00,0x00
};

unsigned char d_ors[] EEMEM = { 
    0x00,0x00,0x04,0x20,0x10,0x00,0x00,0x04,0x20,0x10,
    0x00,0x00,0x40,0x02,0x00,0x00,0x00,0x40,0x02,0x00
};

/* Timing constants for each number of dots lit, in microseconds */

unsigned char times[] EEMEM = {
    105, 35, 55, 78, 105, 134, 167, 202
};

#ifdef SCROLL_MESSAGE

unsigned char msg[] PROGMEM = 
    "       Four 5x7 LED matrix displays with a single ATtiny2313";
    
//unsigned char msg2[] EEMEM = 
//    "       Some further stuff in EEPROM";
 
register char o asm("r3");
register unsigned short i asm("r8");
 
#endif // SCROLL_MESSAGE

#ifdef BRIGHTNESS_CALIBRATION

unsigned short values[] = { 169,169,169,169,169,169,169,169 };
unsigned char  lines=0;

#endif // BRIGHTNESS_CALIBRATION

/* ------------------------- Function Prototypes -------------------------- */

void escape_to_bootloader(void);
void xmit_now(unsigned char c);
void xmit_str_p_now(const prog_char *p);
void handle_display(void);
void clear_bitmap(void);
void msg2bmap(unsigned short pos, unsigned char x0);
unsigned char get_msg_byte(unsigned short pos);
void int2bmap(unsigned short val, unsigned char x0, unsigned char y0);
void putpixel(unsigned char x,unsigned char y);
void scroll_message(void);

/* ----------------------------- Main Program ----------------------------- */

int main(void)
{
    /* If we came from a Watchdog reset, there's probably some bug down there
       that made us lock up or crash, so send us to the bootloader so we can
       upgrade the firmware and hopefully fix it. */
       
    if (MCUSR & _BV(WDRF)) {
        MCUSR &= ~_BV(WDRF);
        wdt_reset();
        escape_to_bootloader();
    }

    /* UART Setup */
    UBRRL  = 25;                     // 19,200 bps with 0.2% error
    UCSRB |= _BV(RXEN) | _BV(TXEN)  // Enable Transmitter and Receiver
          |  _BV(RXCIE);            // And receive interrupts to wake us up
    
#ifdef HAVE_FRAME_FROM_TIMER
    // Timer0 runs in normal mode
    TCCR0B = 4;   // TCLK = CLKIO/1024 = 8MHz/1024 = 7,182.5 Hz    

    // USI setup
    USICR = _BV(USICS0) | _BV(USIOIE);
    USISR = 10;
#endif // HAVE_FRAME_FROM_TIMER

    // Timer1 setup
    TCCR1A = 0;
    TCCR1B = _BV(WGM12)             // CTC mode
           | 2;                     // CLK/64 = 1MHz
    OCR1A  = 124;                   // Base frequency = 8KHz
    TIMSK  = _BV(OCIE1A);

    MCUCR |= _BV(PUD);
    UCSRB &= ~_BV(TXEN);             // Disable transmitter
    sei();
    o = 0;
    i = 0;
    current = bitmap1;
    bitmap  = bitmap2;
    for (;;) {
        wdt_reset();
        set_sleep_mode(SLEEP_MODE_IDLE);
        sleep_mode();
        asm("nop");
        if (GPIOR0 & _BV(FLAG_HAS_INTERRUPT)) {
            GPIOR0 &= ~_BV(FLAG_HAS_INTERRUPT);
            handle_display();
        }
        if (GPIOR0 & _BV(FLAG_HAS_FRAME)) {
            GPIOR0 &= ~_BV(FLAG_HAS_FRAME);
#ifdef SCROLL_MESSAGE
            if (!(GPIOR0 & _BV(FLAG_SINGLE_STEP))) {
                scroll_message();
            }
#endif // SCROLL_MESSAGE
#ifdef BRIGHTNESS_CALIBRATION
            clear_bitmap();
            int2bmap(values[lines],13,0);
            unsigned char n,m;
            if (lines>2) {
                for (n=0;n<lines-1;n++) {
                    putpixel(n,0);
                    putpixel(n,1);
                }
            }
            for (n=0;n<lines;n++) {
                for (m=2;m<5;m++) putpixel(n,m);
            }
#endif // BRIGHTNESS_CALIBRATION
            GPIOR0 |= _BV(FLAG_SWITCH_FRAME);
        }
        if (UCSRA & _BV(RXC)) {               // Have we received something?
            unsigned char c=UDR;             // If so, get it
            if (c==3) {                      // CTRL-C jumps to bootloader
                escape_to_bootloader();
            } 
#ifdef HAVE_DOT_TEST
            else if (c==13) {
                clear_bitmap();
                putpixel(x,y);
                if (++x==28) {
                    x=0;
                    if (++y==5) y=0;
                }
                GPIOR0 |= _BV(FLAG_SWITCH_FRAME);
            } 
#endif // HAVE_DOT_TEST
#ifdef BRIGHTNESS_CALIBRATION
            else if (c=='z') {
                if (lines==0) lines=7; else lines--;
            } 
            else if (c=='x') {
                if (lines==7) lines=0; else lines++;
            } 
            else if (c==',') {
                if (!values[lines]) values[lines]=9999; else values[lines]--;
            } else if (c=='.') {
                if (values[lines]==9999) values[lines]=0; else values[lines]++;
            }
#endif // BRIGHTNESS_CALIBRATION
#ifdef SCROLL_MESSAGE_SINGLE_STEP
            else if (c==' ') {
                if (GPIOR0 & _BV(FLAG_SINGLE_STEP)) {
                    scroll_message();
                }
            } 
            else if (c=='s') {
                GPIOR0 ^= _BV(FLAG_SINGLE_STEP);
            } 
#endif // SCROLL_MESSAGE_SINGLE_STEP
        }
    }
}

/* --------------------------- Utility Routines --------------------------- */

void clear_bitmap(void)
{
    unsigned char n;
    for (n=0;n<20;n++) bitmap[n]=0;
}

#ifdef SCROLL_MESSAGE

// This was originally meant to support parts of the scrolling message
// to be both in flash and in EEPROM

unsigned char get_msg_byte(unsigned short pos)
{
    //if (pos>=sizeof(msg)-1) {
    //    return eeprom_read_byte(msg2+pos-sizeof(msg)-1);
    //} else {
        return pgm_read_byte(msg+pos);
    //}
}

void scroll_message(void)
{
    clear_bitmap();
    if (--o==-4) {
        o=0;
        if (!get_msg_byte(++i)) i=0;
    }
    msg2bmap(i,o);
    GPIOR0 |= _BV(FLAG_SWITCH_FRAME);
}

// Prints a message in the framebuffer, converting characters in their
// pixel representations according to the font table. 'pos' specifies
// the beginning of the message, x0 means the start 'x' coordinate on
// the bitmap. Given that the framebuffer can hold at most seven characters,
// this routine stops after eight characters regardless of whether the
// string has actually ended or not.

void msg2bmap(unsigned short pos, unsigned char x0)
{
    x0+=3;
    unsigned char n;
    for (n=0;n<8;n++) {
        unsigned char src_address = get_msg_byte(pos+n);
        if (!src_address) break;
        src_address -= 32;
        unsigned char  src_nibble = src_address&1;
        src_address >>= 1;
        unsigned char save = src_address;
        src_address <<= 2;
        src_address += save;
        unsigned char m;
        for (m=0;m<5;m++) {
            unsigned char data=pgm_read_byte(font+src_address++);
            if (src_nibble) {
                data &= 0x0F;
            } else {
                data >>= 4;
            }
            unsigned char x;
            for (x=0;x<4;x++) {
                if (data&1) putpixel(x0-x,m);
                data>>=1;
            }
        }
        x0+=4;
    }
}

#endif // SCROLL_MESSAGE


#ifdef BRIGHTNESS_CALIBRATION
unsigned short pow10[] PROGMEM = { 1000,100,10,1 };

void int2bmap(unsigned short val, unsigned char x0, unsigned char y0)
{
    x0+=3;
    unsigned char n;
    for (n=0;n<4;n++) {
        unsigned char i;
        unsigned short pt=pgm_read_word(pow10+n);
        for (i=0;;i++) {
            if (val<pt) break;
            val-=pt;
        }
        unsigned char src_address = i+16;
        unsigned char  src_nibble = src_address&1;
        src_address >>= 1;
        unsigned char save = src_address;
        src_address <<= 2;
        src_address += save;
        unsigned char m;
        for (m=0;m<5;m++) {
            unsigned char data=pgm_read_byte(font+src_address++);
            if (src_nibble) {
                data &= 0x0F;
            } else {
                data >>= 4;
            }
            unsigned char x;
            for (x=0;x<4;x++) {
                if (data&1) putpixel(x0-x,y0+m);
                data>>=1;
            }
        }
        x0+=4;
    }
}

#endif // BRIGHTNESS_CALIBRATION

void putpixel(unsigned char x,unsigned char y)
{
    if (x>=28 || y>=5) return;
    unsigned char address;
    for (address=0;x>=7;x-=7) address++;
    unsigned char save = address;
    address <<= 2;        
    address += save + y;
    bitmap[address] |= (1<<(6-x));
}

/* This is the main multiplexing routine. It performs the necessary 
   bitswapping to get the display data to the appropriate ports, counts
   how many bits are on, and, based on that information, computes the
   appropriate delay to give the impression of a more-or-less even brightness
   distribution. It then computes the final direction flags for each
   display bank, outputs the data and increments the frame counter.
   After the end of the refresh cycle, it swaps the framebuffers.
   
   The timing calculations needs refinement, particularly for the PA2 case:
   when single-stepping it is easy to see that the last rows of the last
   two displays (precisely the ones controlled by PA2) differ in brightness
   of the rest. The other rows are reasonably uniform, but the global
   brightness seems to vary depending of the number of pixels lit.
   */

void handle_display(void)
{
    GPIOR1=current[cycle];
    unsigned char a=0,b=0,d=0,bits=0;
    if (GPIOR1 & _BV(0)) { a |= _BV(1); bits++; }
    if (GPIOR1 & _BV(1)) { a |= _BV(0); bits++; }
    if (GPIOR1 & _BV(2)) { b |= _BV(4); bits++; }
    if (GPIOR1 & _BV(3)) { d |= _BV(3); bits++; }
    if (GPIOR1 & _BV(4)) { b |= _BV(2); bits++; }
    if (GPIOR1 & _BV(5)) { b |= _BV(1); bits++; }
    if (GPIOR1 & _BV(6)) { b |= _BV(0); bits++; }
        
    unsigned char da = a;
    unsigned char db = eeprom_read_byte(b_ors+cycle) | b;
    unsigned char dd = eeprom_read_byte(d_ors+cycle) | d;

    OCR1A = eeprom_read_byte(times+bits);
    if (cycle<5) {
        //OCR1A = values[bits];
    } else if (cycle<10) {
        //OCR1A = 191;
        a ^= 0x03;
        b ^= 0x3F;
        d ^= 0x3C;
    } else if (cycle<15) {
        if (cycle==14) {
            da |= _BV(PA2);
            OCR1A = ((unsigned short)(bits+1)<<7);
        } else {
           //OCR1A = 149;
        }
    } else {
        a ^= 0x07;
        b ^= 0xD7;
        d ^= 0x4A;
        if (cycle==19) {
            da |= _BV(PA2);
            if (bits>3) bits+=2;
            OCR1A=((unsigned short)(bits+2)<<9);
        } else {
            //OCR1A = 124;
        }
    }
    PORTA = a; DDRA = da;
    PORTB = b; DDRB = db;
    PORTD = d; DDRD = dd;
    if (++cycle>=20) {
        cycle=0;
        if (GPIOR0 & _BV(FLAG_SWITCH_FRAME)) {
            unsigned char *aux;
            aux = current;
            current = bitmap;
            bitmap = current;
            GPIOR0 &= ~_BV(FLAG_SWITCH_FRAME);
        }
#ifndef HAVE_FRAME_FROM_TIMER
        GPIOR0 |= _BV(FLAG_HAS_FRAME);
        if (++frame==18) {
            frame=0;
        }
#endif // !HAVE_FRAME_FROM_TIMER
    }
}

void escape_to_bootloader(void)
{
    cli();                       // Globally disable interrupts
    DDRA = PORTA = 0;
    DDRB = PORTB = 0;
    DDRD = PORTD = 0;
    // Jump to bootloader
    asm("rjmp bootstart");
}

/* -------------------------- Interrupt Handlers -------------------------- */

void SIG_TIMER1_COMPA(void) __attribute__((naked));
void SIG_TIMER1_COMPA(void)
{
    GPIOR0 |= _BV(FLAG_HAS_INTERRUPT);
    asm("reti");
}

#ifdef HAVE_FRAME_FROM_TIMER

SIGNAL(SIG_USI_OVERFLOW)
{
    // Unlike most interrupt flags, USI Overflow Interrupt isn't automatically
    // cleared when its interrupt vector is executed, so we must do it by
    // ourselves by writing 1 to that bit.
    USISR = 10 | _BV(USIOIF);
    GPIOR0 |= _BV(FLAG_HAS_FRAME);
}

#endif // HAVE_FRAME_FROM_TIMER

// The USART data receive interrupt is empty; it is enabled just to wake us
// up from sleep mode

EMPTY_INTERRUPT(SIG_USART0_RECV);
